import os

from subprocess import check_call

import shutil

from hatchling.builders.hooks.plugin.interface import BuildHookInterface


here = os.path.abspath(os.path.dirname(__file__))


def build_javascript():
    check_call(
        ['yarn', 'install'],
        cwd=here,
    )
    check_call(
        ['yarn', 'build:webpack'],
        cwd=here,
    )
    shutil.copyfile(
        './dist/main.jupyter-functions.js',
        './jupyter_functions/static/main.jupyter-functions.js'
    )


class JupyterBuildHook(BuildHookInterface):
    def initialize(self, version, build_data):
        if self.target_name == 'editable':
            build_javascript()
        elif self.target_name == 'wheel':
            build_javascript()
        elif self.target_name == 'sdist':
            build_javascript()



"""WebSocket Relay Handler."""

from tornado.websocket import WebSocketHandler

from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.base.zmqhandlers import WebSocketMixin


CONNECTED = set()


class WsRelayHandler(WebSocketMixin, WebSocketHandler, JupyterHandler):
    """WebSocket Relay Handler"""

    def open(self, *args, **kwargs):
        """WsRelayHandler open"""
        self.log.info("WsRelayHandler opened.")
        super(WebSocketMixin, self).open(*args, **kwargs)
        CONNECTED.add(self)

    def on_message(self, message):
        """WsRelayHandler on_message"""
        self.log.info("WsRelayHandler message: " + message)
        peers = {peer for peer in CONNECTED if peer is not self}
        for peer in peers:
            peer.write_message(str(message))

    def on_close(self):
        """WsRelayHandler on_close"""
        self.log.info("WsRelayHandler closed.")
        CONNECTED.remove(self)

    # CORS
#    def check_origin(self, origin):
#        return True

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header('Access-Control-Allow-Methods', 'POST, PUT, DELETE, GET, OPTIONS')
        self.set_header("Access-Control-Allow-Credentials", "true")
        self.set_header("Access-Control-Allow-Headers", "Authorization, Content-Type, Depth, User-Agent, X-File-Size, X-Requested-With, X-Requested-By, If-Modified-Since, X-File-Name, Cache-Control")

    def options(self):
        self.set_status(204)
        self.finish()



"""Proxy Handler."""

from tornado.ioloop import IOLoop
from tornado.websocket import WebSocketHandler, websocket_connect

from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.extension.handler import ExtensionHandlerMixin
from jupyter_server.base.zmqhandlers import WebSocketMixin


# pylint: disable=W0223
class DefaultHandler(ExtensionHandlerMixin, JupyterHandler):

    def get(self):
        # The name of the extension to which this handler is linked.
        self.log.info("Extension Name in {} Default Handler: {}".format(self.name, self.name))
        # A method for getting the url to static files (prefixed with /static/<name>).
        self.log.info("Static URL for / in datalayer_rtc Default Handler:".format(self.static_url(path='/')))
        self.write('<h1>Hello Simple 1 - I am the default...</h1>')
        self.write('Config in {} Default Handler: {}'.format(self.name, self.config))


class WsProxyHandler(WebSocketMixin, WebSocketHandler, ExtensionHandlerMixin, JupyterHandler):

    ws_cnx = {}

    def ws_key(self, port, doc):
        return f'{port}-{doc}'

    async def open(self):
        port = self.get_argument('port', default=None)
        doc = self.get_argument('doc', default=None)
        ws_key = self.ws_key(port, doc)
        print(f"WebSocket Proxy {ws_key}, open")
        print(f"\nDEBUG {self.request}, {self.request.remote_ip}  \n")
        if ws_key in self.ws_cnx:
            return
        self.ws_cnx[ws_key] = await websocket_connect(f'ws://localhost:{port}/{doc}')
        async def proxy_loop():
            while True:
                msg = await self.ws_cnx[ws_key].read_message()
                if msg is None:
                    break
                # print("proxy loop msg : ", msg)
                print(f"Websocket Proxy proxy loop msg: {ws_key}")
                await self.write_message(msg, binary = True)
        IOLoop.current().spawn_callback(proxy_loop)

    def on_message(self, message,  *args, **kwargs):
        port = self.get_argument('port', default=None)
        doc = self.get_argument('doc', default=None)
        ws_key = self.ws_key(port, doc)
        print(f"WebSocket Proxy {ws_key}")
        self.ws_cnx[ws_key].write_message(message, binary = True)

    def on_close(self,  *args, **kwargs):
        port = self.get_argument('port', default=None)
        doc = self.get_argument('doc', default=None)
        ws_key = self.ws_key(port, doc)
        print(f"WebSocket Proxy {ws_key}, on close")
        if ws_key in self.ws_cnx:
            self.ws_cnx[ws_key].close()
            print(f"WebSocket Proxy {ws_key} closed")
            del self.ws_cnx[ws_key]

    # CORS
#    def check_origin(self, origin):
#        return True

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header('Access-Control-Allow-Methods', 'POST, PUT, DELETE, GET, OPTIONS')
        self.set_header("Access-Control-Allow-Credentials", "true")
        self.set_header("Access-Control-Allow-Headers", "Authorization, Content-Type, Depth, User-Agent, X-File-Size, X-Requested-With, X-Requested-By, If-Modified-Since, X-File-Name, Cache-Control")

    def options(self):
        self.set_status(204)
        self.finish()



from typing import Any, Dict, List

from ._version import __version__
from .serverapplication import JupyterFunctionsExtensionApp


def _jupyter_server_extension_points() -> List[Dict[str, Any]]:
    return [{
        "module": "jupyter_functions",
        "app": JupyterFunctionsExtensionApp,
    }]


def _jupyter_labextension_paths() -> List[Dict[str, str]]:
    return [{
        "src": "labextension",
        "dest": "@datalayer/jupyter-functions"
    }]



"""Config handler."""

import json

import tornado

from jupyter_server.base.handlers import APIHandler
from jupyter_server.extension.handler import ExtensionHandlerMixin

from ..._version import __version__


# pylint: disable=W0223
class ConfigHandler(ExtensionHandlerMixin, APIHandler):
    """The handler for configurations."""

    @tornado.web.authenticated
    def get(self):
        """Returns the configurations of the server extensions."""
        res = json.dumps({
            "extension": "jupyter_functions",
            "version": __version__
        })
        self.finish(res)



"""WebSocket ping handler."""

from tornado import ioloop
from tornado.websocket import WebSocketHandler

from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.base.zmqhandlers import WebSocketMixin


# pylint: disable=W0223
class WsPingHandler(WebSocketMixin, WebSocketHandler, JupyterHandler):
    """WebSocket Pong handler"""

    def open(self, *args, **kwargs):
        """open"""
        print("WebSocket opened.")
        super(WebSocketMixin, self).open(*args, **kwargs)
        loop = ioloop.IOLoop.current()
        self.last_ping = loop.time()
        self.last_pong = self.last_ping
        self.ping_callback = ioloop.PeriodicCallback(
            self.send_hello,
            1000,
        )
        self.ping_callback.start()

    def send_hello(self):
        """Send a hello."""
        if self.ws_connection is None and self.ping_callback is not None:
            self.ping_callback.stop()
            return
        if self.ws_connection.client_terminated:
            self.close()
            return
        # check for timeout on pong.  Make sure that we really have sent a recent ping in
        # case the machine with both server and client has been suspended since the last ping.
        now = ioloop.IOLoop.current().time()
        since_last_hello = 1e3 * (now - self.last_pong)
        since_last_hello = 1e3 * (now - self.last_ping)
        if since_last_hello < 2 * self.ping_interval and since_last_hello > self.ping_timeout:
            self.log.warning("WebSocket ping timeout after %i ms.", since_last_hello)
            self.close()
            return
        self.last_ping = now
        self.write_message('hello...')


    def on_message(self, message):
        """on_message"""
        print("WebSocket message: " + message)
        self.write_message(str(message) + '... pong')

    def on_close(self):
        """on_close"""
        print("WebSocket closed")



"""The Jupyter Functions Server application."""

import os

from traitlets import Unicode

from jupyter_server.utils import url_path_join
from jupyter_server.extension.application import ExtensionApp, ExtensionAppJinjaMixin

from ._version import __version__

from .handlers.index.handler import IndexHandler
from .handlers.config.handler import ConfigHandler
from .handlers.echo.handler import WsEchoHandler
from .handlers.relay.handler import WsRelayHandler
from .handlers.proxy.handler import WsProxyHandler
from .handlers.ping.handler import WsPingHandler


DEFAULT_STATIC_FILES_PATH = os.path.join(os.path.dirname(__file__), "./static")

DEFAULT_TEMPLATE_FILES_PATH = os.path.join(os.path.dirname(__file__), "./templates")


class JupyterFunctionsExtensionApp(ExtensionAppJinjaMixin, ExtensionApp):
    """The Jupyter Functions Server extension."""

    name = "jupyter_functions"

    extension_url = "/jupyter_functions"

    load_other_extensions = True

    static_paths = [DEFAULT_STATIC_FILES_PATH]
    template_paths = [DEFAULT_TEMPLATE_FILES_PATH]

    config_a = Unicode("", config=True, help="Config A example.")
    config_b = Unicode("", config=True, help="Config B example.")
    config_c = Unicode("", config=True, help="Config C example.")

    def initialize_settings(self):
        self.log.debug("Jupyter Functions Config {}".format(self.config))

    def initialize_templates(self):
        """
        from jinja2 import Environment, PackageLoader, FileSystemLoader, meta
        env = Environment(loader = FileSystemLoader(self.template_paths))
        template_settings = {f'{self.name}_jinja2_env': env}
        self.settings.update(**template_settings)
        env = Environment(loader=PackageLoader('jupyter_functions', 'templates'))
        template_source = env.loader.get_source(env, 'index.html')
        parsed_content = env.parse(template_source)
        variables = meta.find_undeclared_variables(parsed_content)
        ref_templates = meta.find_referenced_templates(parsed_content)
        """
        self.serverapp.jinja_template_vars.update({"jupyter_functions_version" : __version__})

    def initialize_handlers(self):
        self.log.debug("Jupyter Functions Jinja2 Env {}".format(self.settings['jupyter_functions_jinja2_env']))
#        host_pattern = ".*$"
#        base_url = web_app.settings["base_url"]
#        route_pattern = url_path_join(base_url, "jupyter_functions", "get_example")
#        echo_pattern = url_path_join(base_url, "jupyter_functions", "echo")
#        handlers = [
#            (route_pattern, JupyterExecHandler),
#            (echo_pattern, WsEchoHandler),
#        ]
        handlers = [
            ("jupyter_functions", IndexHandler),
            (url_path_join("jupyter_functions", "config"), ConfigHandler),
            (url_path_join("jupyter_functions", "echo"), WsEchoHandler),
            (url_path_join("jupyter_functions", "relay"), WsRelayHandler),
            (url_path_join("jupyter_functions", "proxy"), WsProxyHandler),
            (url_path_join("jupyter_functions", "ping"), WsPingHandler),
        ]
        self.handlers.extend(handlers)


# -----------------------------------------------------------------------------
# Main entry point
# -----------------------------------------------------------------------------

main = launch_new_instance = JupyterFunctionsExtensionApp.launch_instance



"""WebSocket echo handler."""

from tornado import ioloop
from tornado.websocket import WebSocketHandler

from jupyter_server.base.handlers import JupyterHandler
from jupyter_server.base.zmqhandlers import WebSocketMixin


# pylint: disable=W0223
class WsEchoHandler(WebSocketMixin, WebSocketHandler, JupyterHandler):
    """WebScoekt echo handler."""

    def _send_hello(self):
        """Send a hello."""
        if self.ws_connection is None and self.ping_callback is not None:
            self.ping_callback.stop()
            return
        if self.ws_connection.client_terminated:
            self.close()
            return
        # check for timeout on pong.  Make sure that we really have sent a recent ping in
        # case the machine with both server and client has been suspended since the last ping.
        now = ioloop.IOLoop.current().time()
        since_last_hello = 1e3 * (now - self.last_pong)
        since_last_hello = 1e3 * (now - self.last_ping)
        if since_last_hello < 2 * self.ping_interval and since_last_hello > self.ping_timeout:
            self.log.warning("WebSocket ping timeout after %i ms.", since_last_hello)
            self.close()
            return
        self.last_ping = now
        self.write_message('hello...')


    def open(self, *args, **kwargs):
        """WebSocket open"""
        self.log.info("WebSocket opened.")
        super(WebSocketMixin, self).open(*args, **kwargs)
        loop = ioloop.IOLoop.current()
        self.last_ping = loop.time()
        self.last_pong = self.last_ping
        self.ping_callback = ioloop.PeriodicCallback(
            self._send_hello,
            1000,
        )
        self.ping_callback.start()

    def on_message(self, message):
        """WebSocket on message"""
        self.log.info("WebSocket message: " + message)
        self.write_message(str(message) + '... pong')

    def on_close(self):
        """WebSocket on close"""
        self.log.info("WebSocket closed")

    # CORS

#    def check_origin(self, origin):
#        return True

    def set_default_headers(self):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header('Access-Control-Allow-Methods', 'POST, PUT, DELETE, GET, OPTIONS')
        self.set_header("Access-Control-Allow-Credentials", "true")
        self.set_header("Access-Control-Allow-Headers", "Authorization, Content-Type, Depth, User-Agent, X-File-Size, X-Requested-With, X-Requested-By, If-Modified-Since, X-File-Name, Cache-Control")

    def options(self):
        self.set_status(204)
        self.finish()
